/* $Id: addrlib.h,v 2.11 2011-10-27 16:34:55 hurtta Exp $ */

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.11 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> (was hurtta+elm@posti.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

/* Defines from elm_defs.h ------------------------------------------------ */


/** "exit_disposition" values */
enum exit_disposition {
    UNSET =	0,
    KEEP  =	1,
    STORE =	2,
    DELETE =	3,
    HIDE =    4
};


/* Content-Disposition */
enum {
    DISP_INLINE	     = 0,
    DISP_ATTACH	     = 1,
    DISP_AUTOATTACH  = 2
};

enum encoding {
    ENCODING_ILLEGAL =	-1,
    ENCODING_NONE    =	0,
    ENCODING_7BIT    =	1,
    ENCODING_8BIT    =	2,
    ENCODING_BINARY  =	3,
    ENCODING_QUOTED  =	4,
    ENCODING_BASE64  =	5,
    ENCODING_EXPERIMENTAL  =	6,
    ENCODING_UUENCODED =  7 };


/* XXX struct mimeinfo is embedded to struct header_rec */
#define MIME_magic              0xFD00
typedef struct mimeinfo {
    unsigned short magic;         /* MIME_magic */
    unsigned int disposition : 2; /* DISP_INLINE or DISP_ATTACH or
				     DISP_AUTOATTACH
				  */

    unsigned int mime_flags : 5;  /* NOTPLAIN_need_metamail           0x01
				     NOTPLAIN_need_mailcap            0x02
				     NOTPLAIN_canuse_mailcap          0x04
				     NOTPLAIN_is_fallback             0x08
				     NOTPLAIN_metamail_blacklisted    0x10
				  */

    unsigned int unlink : 2;     /* used to mark that the file should be
				  * unlinked after sending the attachment 
				  */

    struct media_type *     TYPE;
    struct mime_param *     TYPE_opts;
    struct mime_param *     DISPOSITION_opts;
    struct string     *     description;
    
    enum encoding encoding;      /* Notice: ENCODING_ILLEGAL == -1 */
    long length;
    long begin_offset; /* the beginning of the attachment (includes headers) */
    long offset;       /* where the real data starts */    

    struct mime_parser_data         * parser_data; 

    struct mime_selected_handler    * handler_data;   

    struct metamail_mailcap_entry   * metamail_blacklisted;


    /* These next fields are only used when sending multipart messages. */
    char          * pathname0;
    struct string * dispname; 

} mime_t;

struct header_rec {
    int  lines;		/** # of lines in the message	**/
    int  status;		/** Urgent, Deleted, Expired?	**/
    int  status1;		/** Flagged?	                **/
    int  index_number_X;	/** relative loc in file...	**/
    int  encrypted;		/** whether msg has encryption	**/
    enum exit_disposition   exit_disposition;	/** whether to keep, store, delete **/
    int  status_chgd;	         /** whether became read or old, etc. **/
    long content_length;	/** content_length in bytes from message header	**/
    long offset;		/** offset in bytes of message	**/
    time_t received_time;	/** when elm received here	**/
    char env_from[STRING];      /* sender from "From "          */

    struct addr_list *from;	/** who sent the message?	**/
    struct addr_list *to;	/** who it was sent to		**/
    struct addr_list *cc;
    struct addr_list *reply_to;

    struct message_id * message_id;	/** the Message-ID: value	**/
    struct references * in_reply_to;    /* In-Reply-To   */
    struct references * references;     /* References */

    char time_zone[12];	        /**                incl. tz	**/
    time_t time_sent;	        /** gmt when sent for sorting	**/
    char time_menu[SHORT];	/** just the month..day for menu **/
    time_t tz_offset;	        /** offset to gmt of time sent	**/
    struct string *subject;     /** The subject of the mail	**/

    char mailx_status[WLEN];    /** mailx status flags (RO...)	**/
    mime_t mime_rec;            /** MIME information for this message **/


    /* == charset_t, but it is not yet declared ... */
    struct charcode_info *  header_charset;  /* charset for headers ...    */
    struct charcode_info *  override_charset; /* For O)verride command */

    struct mbx_hdr_info    * mbx_info;

    struct partial_vector  * partial_vector;
    int                      partial_len;

    struct list_info       * list_info; 
    struct header_errors   * header_error;

#ifdef USE_PGP
    unsigned int pgp : 3;
#endif
    unsigned int binary :1;  /* Binary message -- headers are terminated
			      * with \r\n 
			      */
    unsigned int have_from :1;  /* Body have unescaped "From " */
    unsigned int body_parsed :1;  /* Whatever body of message is read */
    unsigned int mime_parsed :1;  /* Whatever mime structure is parsed */
};




/* ========================================================================= */

struct address;

/* address list operations ------------------------------------------------- */

extern struct addr_list  * new_empty_addr_list P_((int alloced));

extern void free_addr_list P_((struct addr_list **list));

extern struct addr_list * parse_header_address P_((const char *headername,
						   const char *buffer, 
						   int demime, 
						   charset_t defcharset,
						   struct header_errors **header_error));

extern struct addr_list * dup_addr_list P_((const struct addr_list *list));

extern void append_addr_list P_((struct addr_list *list,
				 const struct addr_list *src));


extern int addr_list_item_count   P_((const struct addr_list *list));
extern int addr_list_group_count  P_((const struct addr_list *list));

/* *group_num == -1 if not part of group */
extern const struct address * addr_list_get_item P_((const struct addr_list 
						     *list,
						     int idx,
						     int *group_num));
extern const struct string * addr_list_get_group P_((const struct addr_list 
						     *list,
						     int idx));


extern void addr_list_replace_item P_((struct addr_list *list,
				       int idx,
				       const struct address *addr,
				       int group_num));

/* groupidx == -1 if not part of group */
extern int add_address_to_list P_((struct addr_list *list,
				   const struct address *addr,
				   int groupidx));

/* Returns groupidx which can be used with add_address_to_list */
extern int add_group_to_list P_((struct addr_list *list,
				 const struct string *phrase));

/* Return -1 if group not found */
extern int addr_list_lookup_group P_((struct addr_list *list,
				      const struct string *group,
				      int create));

/* address operations ------------------------------------------------------ */

extern void free_address P_((struct address **addr));

extern struct address * new_address P_((const char * ascii_addr,
					const struct string  * phrase,
					const struct string  * comment));
extern struct address * new_address_string P_((const struct string * address,
					       const struct string * phrase,
					       const struct string * comment));


extern const struct string * address_get_phrase P_((const struct address *
						    addr));
extern const struct string * address_get_comment P_((const struct address *
						     addr));
extern const char * address_get_ascii_addr P_((const struct address *addr));

extern struct address   * parse_one_address P_((const char *buffer,
						int demime,
						charset_t defcharset));

extern struct address * dup_address P_((const struct address *addr));

extern struct address   * parse_one_string_address P_((const struct string *str));
extern struct addr_list * parse_string_address     P_((const struct string *str));

/* Caller must free string */
extern struct string * address_to_string P_((const struct address *addr));

/* Caller must frees tring */
extern struct string * addr_list_to_string P_((const  struct addr_list *list));

/* Caller must free string */
extern struct string *string_quote_phrase P_((const struct string *phrase));




/* mailing list  ------------------------------------------------------ */

struct mlist_info;
struct mlist_conf;

typedef int mlist_match_id_f P_((char * list_id,
				 struct list_info *list));

extern struct mlist_scan * mlist_get_matches P_((struct mlist_conf *list,
						 struct header_rec *hdr,
						 mlist_match_id_f *f));

extern void free_mlist_scan P_((struct mlist_scan **match));

extern void append_mlist_scan P_((struct mlist_scan *list,
				  const struct mlist_scan *other));

/* List is non-empty */
extern int have_mlist_scan P_((struct mlist_scan *match));

extern const struct mlist_info * get_mlist_info P_((struct mlist_scan *match));

/* Display address --------------------------------------------------------- */

struct header_rec;  
struct string;

typedef void decode_who P_((int class, char *ptr, int size));
extern int DisplayAddress  P_((struct header_rec *h,struct string **f));

/* okay_addr.c ------------------------------------------------------------- */

extern int okay_address_l P_((const struct address  *address, 
			      const struct addr_list *return_address));

/* mailer.c    ------------------------------------------------------------- */

struct mailing_headers;

extern char **argv_from_headers   P_((struct mailing_headers * headers));

extern void free_argv P_((char ***argv));

/* headers.c   ------------------------------------------------------------- */

#define   RHL_MARK_FOLDING      1
#define   RHL_CHECK_HEADER      2

extern int read_header_line             P_((FILE *,char *,int,int));
extern int state_read_hdr_line          P_((struct in_state *,char *, int, int));

typedef struct header_info * header_ptr;

/* Returns pointer to static storage */
extern const char * give_header_name P_((header_ptr X));
extern int is_mailing_header P_((header_ptr X));

extern int add_to_mailing_header P_((struct mailing_headers * headers,
				     header_ptr X,const char *value,
				     int demime,
				     charset_t defcharset,
				     int replace
				     ));

#define HEADER_magic        0xFC00

typedef struct header_list {
    unsigned short magic;
    header_ptr header_name;
    struct header_list * next_this_header;
    struct header_list * next_header;
    struct header_list * next_other_header;
    char * body;
} * header_list_ptr;

extern void update_header_list P_((header_list_ptr *list,
				   header_list_ptr *last1,
				   const char * name,
				   const char * value));

extern header_ptr find_header  P_((const char *name, int create_flag));
extern header_list_ptr locate_header      P_((header_list_ptr,header_ptr));

#define locate_header_by_name(h,n)      locate_header(h,find_header(n,0))
extern void delete_headers              P_((header_list_ptr *hdr));


extern struct string * give_decoded_header P_((header_list_ptr tmphdr,
					       int demime, 
					       charset_t defcharset));

/* outheaders.c ------------------------------------------------------------ */

struct textual {
    struct string *Textual;
    int pos;
    int len;
};

struct expanded_address {
    unsigned short magic;               /* EXP_ADDR_magic */
    struct addr_list  *addrs;           
    
    struct textual *surface;
    int surface_len;
};

extern void dump_expanded_address P_((int debuglevel, const char *text,
				      struct expanded_address expanded));

extern void expanded_address_from_list P_((struct expanded_address *result,
					    const struct addr_list *addr_list));


struct mailing_headers {
    unsigned short magic;                  /* MAIL_HDR_magic */

    struct string *subject;
    struct expanded_address from;

    struct references * in_reply_to;    /* In-Reply-To   */
    struct references * references;     /* References */

    char *expires;
    char *action; 
    char *priority;
    
    char *sender;

    struct expanded_address reply_to; 
    struct expanded_address to;
    struct expanded_address cc;
 
    struct user_header {
	header_ptr      name;
	struct string  *value;
    }  * user_header;
    int  user_header_count;

    struct expanded_address bcc;
    char *precedence;
    char *expires_days;    

    struct mailer_env_from  *env_from;
};

extern void import_mailheaders P_((struct mailing_headers * headers));

extern  int add_expanded_addr_ P_((struct expanded_address *x,
				   const char *ADR, 
				   const struct string *FN,
				   const struct string *COM));
extern int add_expanded_addr0_ P_((struct expanded_address *x,
				   const struct address *ADR,
				   int g));
extern int add_expanded_addr_g P_((struct expanded_address *x,
				   const struct address *ADR,
				   const struct string *group));

extern  int add_expanded_addr_4 P_((struct expanded_address *x,
				    const char *ADR, 
				    const struct string *FN,
				    const struct string *COM,
				    const struct string *group));


extern int add_group_phrase_ P_((struct expanded_address *x,
				 const struct string *GRP));

extern int add_textual_addr_ P_((struct expanded_address *x,
				 const struct string *TEXTUAL, 
				 int POS, int LEN));

#define ADD_EXPANDED(x,ADR,FN,COM)  add_expanded_addr_(&x,ADR,FN,COM)
#define ADD_EXPANDED0(x,ADR)        add_expanded_addr0_(&x,ADR,-1)
#define ADD_EXPANDED_g(x,ADR,G)        add_expanded_addr0_(&x,ADR,G)
#define ADD_EXPANDED_G(x,ADR,G)        add_expanded_addr_g(&x,ADR,G)
#define ADD_EXPANDED4(x,ADR,FN,COM,GRP) add_expanded_addr_4(&x,ADR,FN,COM,GRP)
#define ADD_GROUP(x,GRP)               add_group_phrase_(&x,GRP)
#define ADD_TEXTUAL(x,TEXTUAL,POS,LEN) add_textual_addr_(&x,TEXTUAL,POS,LEN)


extern void zero_expanded_address P_((struct expanded_address *x));
extern void free_expanded_address P_((struct expanded_address *x));
extern void copy_expanded_address P_((struct expanded_address *result,
				      struct expanded_address source));

extern struct string * make_surface_addr P_((const struct address * address));

/* error.c --------------------------------------------------------- */

extern void free_header_errors P_((struct header_errors **error));

extern void process_header_error P_((struct header_errors **error,
				     const char * format, const char *msg, 
				     ...));

extern int get_header_errors_count P_((struct header_errors *error));
extern const struct string * get_header_errors_msg P_((struct header_errors *
						       error, int idx));

extern void print_header_errors P_((struct header_errors *error));


/* message-id operations -------------------------------------------------- */

extern struct message_id  * new_message_id P_((const char * left,
					       const char * domain,
					       const struct string *comment));

extern struct message_id * dup_message_id P_((const struct message_id *id));

extern struct message_id * parse_header_message_id P_((const char *headername,
						       const char *buffer,
						       int demime,
						       charset_t defcharset,
						       struct header_errors 
						       **header_error));

extern int same_message_id P_((const struct message_id *id1,
			       const struct message_id *id2,
			       int ignore_comment));

extern int cmp_message_id P_((const struct message_id *id1,
			      const struct message_id *id2));
			      
extern void free_message_id P_((struct message_id  ** id));


/* Result is malloced,  < > are included to result, no comments */
extern char * get_message_id_ascii_brackets P_((const struct message_id *id));

extern const struct string * message_id_comment P_((const 
						    struct message_id *id));


/* references / in-reply-to operations --------------------------------------- */

extern struct references * dup_references P_((const struct references *id));
extern struct references * dup_references_no_phrase P_((const struct references *id));

extern struct references * parse_header_references P_((const char *headername,
						       const char *buffer,
						       int demime,
						       charset_t defcharset,
						       struct header_errors 
						       **header_error));

extern int same_references P_((const struct references *id1,
			       const struct references *id2));

extern void free_references P_((struct references  ** id));

extern void append_references P_((struct references  * ref,
				  const struct references  * ref1));

extern struct references * new_references P_((const struct message_id  *id,
					      const struct string      *phrase,
					      const struct string      *comment));

extern void references_add_item P_((struct references * ref,
				    const struct message_id  *id,
				    const struct string      *phrase,
				    const struct string      *comment));

extern int references_item_count P_((const struct references * ref));


extern const struct message_id * references_get_message_id P_((
  			          const struct references * ref,
				  int idx));

extern const struct string * references_get_phrase P_((
  			          const struct references * ref,
				  int idx));

extern const struct string * references_get_comment P_((
  			          const struct references * ref,
				  int idx));

extern int header_references_have_phrase P_((const struct references * ref));

/* write_hdr.c ------------------------------------------------------------- */



extern void write_addr_header P_((struct out_state *mailer,
				  char *hdr_name,
				  struct addr_list * addr,
				  enum encoding top_encoding,
				  int enmime,
				  charset_t hdr_charset,
				  int add_group));

#define HAVE_8BIT    1

extern int check_8bit_str P_((const char *str));
extern int check_8bit_string P_((struct string *P));


extern void write_string_header P_((struct out_state *mailer,
				    const char *hdr_name,
				    struct string * P,
				    enum encoding top_encoding,
				    int enmime,
				    charset_t charset));

extern void write_text_header P_((struct out_state *mailer, 
				  const char *hdr_name, 
				  const char * text,
				  enum encoding top_encoding));

/* top_encoding is not required -- also if mailer->EOLN_is_CRLF
   is set, then \r\n is printed
*/
extern void print_EOLN P_((struct out_state *mailer,
			   enum encoding top_encoding));


void write_references_header P_((struct out_state *mailer,
				 const char *hdr_name,
				 struct references * ids,
				 enum encoding top_encoding,
				 int enmime,
				 charset_t hdr_charset));


/* Result is is malloced */
extern char * address_to_str P_((const struct address *address, int enmime,
				 charset_t hdr_charset));
/* Result is is malloced */
char * list_to_str P_((const struct addr_list *list, int enmime,
		       charset_t hdr_charset));

/* Caller must free string */
extern struct string *string_quote_comment P_((const struct string *phrase));


extern void append_quoted_to_string P_((struct string ** ret,
					const struct string * value));

/* Caller must free string */
extern struct string * unquote_string P_((const struct string * value));






/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
