static char rcsid[] = "@(#)$Id: fastmail.c,v 2.9 2011-07-14 13:36:46 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.9 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@siilo.FMI.FI>
 *                       (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************************
 *  Based on Elm 2.4 utils/fastmail.c. That code was following copyright:
 *
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** This program is specifically written for group mailing lists and
    such batch type mail processing.  It does NOT use aliases at all,
    it does NOT read the /etc/password file to find the From: name
    of the user and does NOT expand any addresses.  It is meant 
    purely as a front-end for either /bin/mail or /usr/lib/sendmail
    (according to what is available on the current system).

         **** This program should be used with CAUTION *****

**/


/** The calling sequence for this program is:

	fastmail {args}  [ filename | - ] full-email-address 

   where args could be any (or all) of;

	   -b bcc-list		(Blind carbon copies to)
	   -c cc-list		(carbon copies to)
	   -C comment-line      (Comments:)
	   -d ...		(debug on)
	   -f from 		(from name)
	   -F from-addr		(the actual address to be put in the From: line)
	   -i msg-id            (In-Reply-To: msgid)
	   -r reply-to-address 	(Reply-To:)
	   -R references        (References:)
	   -P precedence        (Precedence:)
	   -s subject 		(subject of message)
**/

#include "def_utils.h"
#include "addrlib.h"
#include "mailerlib.h"
#include "misclib.h"
#include "s_fastmail.h"

DEBUG_VAR(Debug,__FILE__,"util");

#define  temphome	"/tmp/fastmail."


static void usage P_((void));

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

static void mail_sent P_((union any_fd fd, char * title, struct run_state *rs,
 			  int ret, int exit_stat));

static void mail_sent (fd,title,rs,ret,exit_code)
     union any_fd fd; 
     char * title;
     struct run_state *rs;
     int ret; 
     int exit_code;
{    
    printf("%s\n",title);
    fflush(stdout);

    if (ret < 0) 
	lib_error(CATGETS(elm_msg_cat, FastmailSet, FastmailFailSignal,
			  "%.30s fail: Signal?"),
		  get_mailer_path(fd.mail_fd));
    else if (ret > 0) {
	if (rs->save_errno)
	    lib_error(CATGETS(elm_msg_cat, FastmailSet, FastmailFailErrno,
			      "Failed: %.30s: %.40s"),
		      get_mailer_path(fd.mail_fd),
		      strerror(rs->save_errno));
	else if (exit_code) {
	    lib_error(
		      CATGETS(elm_msg_cat, FastmailSet, 
			      FastmailMailerReturnedError,
			      "mailer returned error status %d"), 
		      exit_code);
	} else
	    lib_error(CATGETS(elm_msg_cat, FastmailSet, FastmailMailSent, 
			      "Mail sent!"));
    } else {
	lib_error(CATGETS(elm_msg_cat, FastmailSet, FastmailLostErrno,
			  "%.30s lost: %.40s"),
		  get_mailer_path(fd.mail_fd),
		  strerror(rs->save_errno));
    }
}

static void mailing_message P_((int background));
static void mailing_message(background)
     int background;
{
    if (!background)
	lib_transient(CATGETS(elm_msg_cat, FastmailSet, FastmailSendingMail,
			      "Sending mail..."));
    else
	lib_error(CATGETS(elm_msg_cat, FastmailSet, 
			  FastmailSendingMailBackground,
			  "Sending mail... in background"));    
}

static void simple_add_addr P_((struct expanded_address *expanded,
				struct mailer_info *mailer_info,
				char * addr,
				charset_t defcharset,
				int demime));
static void simple_add_addr(expanded,mailer_info,addr,defcharset,demime)
     struct expanded_address *expanded;
     struct mailer_info *mailer_info;
     char * addr;
     charset_t defcharset;
     int demime;
{
    struct addr_list * TMP = parse_header_address(NULL,addr,demime,defcharset,
						  NULL);
    int j;


    if (TMP) {
	int len = addr_list_item_count(TMP);

	int idx;

	for (idx = 0; idx < len; idx++) {
	    int group = -1;
	    int new_group_idx = -1;
	    const struct address * address = 
		addr_list_get_item(TMP,idx,&group);
	    
	    const char          * addr     = 
		address_get_ascii_addr(address);
	    const struct string * fullname = 
		address_get_phrase(address);
	    const struct string * comment  = 
		address_get_comment(address);
	    
	    const struct string * groupname = NULL;

	    int pos;
	    struct string *s;

	    if (group >= 0) 
		groupname =  addr_list_get_group(TMP,group);
	    	    
	    if (addr && 
		qstrpbrk_c(addr,"!:@") == NULL &&
		mailer_info &&
		query_mailer_info(mailer_info,MI_USE_DOMAIN)) {

		char * addr1              = safe_strdup(addr);
		struct address * address1 = NULL;
		
		DPRINT(Debug,10,(&Debug,
				 "simple_add_addr: Adding domain %s to address %s\n",
				 hostfullname, addr));
		
		
		
		addr1    = strmcat(addr1,"@");
		addr1    = strmcat(addr1, hostfullname);  
		
		address1 = new_address(addr1,fullname,comment);
		
		addr_list_replace_item(TMP,
				       idx,address1,
				       group);   
		
		free_address(&address1);
		free(addr1); addr1 = NULL;
		
		/* Replace invalidates pointers, so reload new pointers */
		
		address   = addr_list_get_item(TMP,idx,&group);
		addr      = address_get_ascii_addr(address);
		fullname  = address_get_phrase(address);
		comment   = address_get_comment(address);	    
	    }
	    
	    pos = add_expanded_addr_g(expanded,address,groupname);
	    s = make_surface_addr(address);
	    add_textual_addr_(expanded,s,pos,1);
	    free_string(&s);
	}

	free_addr_list(&TMP);
    }
}

static char version_buff[NLEN];

static char *mime_encode_names[] = {
    "none", /* Not used. */
    "7bit",
    "8bit",
    "binary",
    "quoted-printable",
    "base64"
};

int main P_((int argc, char *argv[]));
int main(argc, argv)
     int argc;
     char *argv[];
{
    extern char *optarg;
    extern int optind;

    struct mailing_headers mailing_headers;
    
    char *p;
    int  c;

    struct mail_send_state * M = NULL;
    struct out_state *  mailer = NULL;
    FILE                  * F =NULL;
    char ** addrs3 = NULL;

    int exitcode = 1;
    struct mailer_info *mailer_info = NULL;

    struct string * comments     = NULL;
    struct string * from_string  = NULL;
    
    char * filename = NULL;
    char * tempname = NULL;

    char title[STRING];

    struct scan_list * scanlist = NULL;
    char *ext = NULL;
    
    struct mime_types_item * type_item = NULL;
    
    int need_enc = 0;
    int message_encoding = ENCODING_NONE;
    int an = give_dt_enumerate_as_int(&allow_no_encoding);
    
    media_type_t content_type = NULL;
    char **      ctype_params = NULL;
    int          ctype_param_count = 0;
    charset_t    ctype_charset = NULL;
    
    int is_text = 1;  /* -1 structured, 0 binary, 1 text */
    int test_utf8 = 0;

#if DEBUG
    init_debugfile("FASTMAIL");
#endif
    locale_init();

#if DEBUG
    /* So that is is possible to initialize debugging before all other stuff */
    if (argc > 1 && 0 == strncmp("-d",argv[1],2) &&
	argv[1][2]) {
	set_debugging(argv[1]+2);	   
    }
#endif    

    zero_mailing_headers(&mailing_headers);
    
    user_init();
    init_misclib();
    init_mailerlib();

    init_defaults();

    read_rc_file(0);

    mailer_info   = get_mailer_info();

    ctype_charset = display_charset;

    if (!mailer_info)
	exit(1);

    /* free_mailing_headers will free this */
    if (!mailing_headers.env_from)
	mailing_headers.env_from = mailer_get_env_from(mailer_info);

    /* Read .mailheaders for Return-Path (for envelope sender actually) 
       and user deifined headers
    */
    import_mailheaders(&mailing_headers);

    
    /* FIXME: what is correct character set? */
    if ((p = getenv("REPLYTO")) != NULL)
	simple_add_addr(&mailing_headers.reply_to,
			mailer_info,
			p,
			display_charset,
			!allow_no_hdrencoding); 
    
    /* auto bcc line */
    if ((p = getenv("BCC")) != NULL)   
	simple_add_addr(&mailing_headers.bcc,
			mailer_info,
			p,
			display_charset,
			!allow_no_hdrencoding); 
    
    while ((c = getopt(argc, argv, "b:c:C:d:f:F:i:r:R:P:s:t:")) != EOF) {
	switch (c) {
	case 'b' : 
	    
	    simple_add_addr(&mailing_headers.bcc,mailer_info,
			    optarg,system_charset,
			    !allow_no_hdrencoding); 
	    break;

	case 'c' : 
	    simple_add_addr(&mailing_headers.cc,mailer_info,
			    optarg,system_charset,
			    !allow_no_hdrencoding); 
	    break;

	case 'C' : 
	    
	    if ( comments )
		usage();
	    
	    comments = hdr_to_string(HDR_TEXT,optarg,system_charset,
				     !allow_no_hdrencoding);
	    break;

	case 'd' : 
#if DEBUG
	    set_debugging(optarg);	   
#else
	    lib_error(CATGETS(elm_msg_cat, FastmailSet, 
			      FastmailIngoringDebug,
			      "Warning: system created without debugging enabled - request ignored\n"));
#endif
	    break;
	case 'f' : 
	    from_string = hdr_to_string(HDR_TEXT,optarg,system_charset,
					!allow_no_hdrencoding);
			     	
	    break;
	case 'F' : 
	    simple_add_addr(&mailing_headers.from,mailer_info,
			    optarg,system_charset,
			    !allow_no_hdrencoding); 
	    break;
	case 'i' : 
	    if (mailing_headers.in_reply_to)
		usage();

	    mailing_headers.in_reply_to =
		parse_header_references(NULL,optarg,!allow_no_hdrencoding,
					system_charset,NULL);

	    if (!add_irt_phrase && header_references_have_phrase(mailing_headers.in_reply_to)) {

		lib_error(CATGETS(elm_msg_cat, FastmailSet, 
				  FastmailHdrPhraseIRT,
				  "In-reply-to header should include only message-id"));
	    
	    }

					
	    
	    break;

	case 'r' : 
	    simple_add_addr(&mailing_headers.reply_to,mailer_info,
			    optarg,system_charset,
			    !allow_no_hdrencoding); 
	    break;
	case 'R' : 
	    if (mailing_headers.references)
		usage();

	    mailing_headers.references =
		parse_header_references(NULL,optarg,!allow_no_hdrencoding,
					system_charset,NULL);
	    
	    if (header_references_have_phrase(mailing_headers.references)) {

		lib_error(CATGETS(elm_msg_cat, FastmailSet, 
				  FastmailHdrPhraseReferences,
				  "References header should include only one or more message-ids"));

	    }


	    break;
	case 'P' : 
	    if (mailing_headers.precedence)
		usage();
	    
	    if (check_8bit_str(optarg)) {
		lib_error(CATGETS(elm_msg_cat, FastmailSet, 
				  FastmailNo8bitonP,
				  "8-bit data not allowed on -P: %s"),
			  optarg);
		
		goto FAILURE;
	    } 

	    mailing_headers.precedence = safe_strdup(optarg);
	    break;

	case 's' : 
	    if (mailing_headers.subject)
		usage();
	    mailing_headers.subject = 
		hdr_to_string(HDR_TEXT,optarg,system_charset,
			      !allow_no_hdrencoding);
	    
	    break;

 	case 't': {
 	    char * X;
 	    char * walk = NULL;
 	    char * type;
	    
 	    if (check_8bit_str(optarg)) {
 		lib_error(CATGETS(elm_msg_cat, FastmailSet, 
 				  FastmailNo8bitont,
 				  "8-bit data not allowed on -t: %s"),
 			  optarg);
		
 		goto FAILURE;
 	    } 
	    
 	    X = safe_strdup(optarg);
 	    type = mime_parse_content_opts(X,&walk);
	    
 	    if (type) {
 		char * A;
 		char * p;
 		char * subtype;
		
 		rfc822_reap_comments (type, NULL, 0);
		
 		/* Parse type */
		
 		while (*type && whitespace(*type))
 		    type++;
		
 		if (!*type)
 		    goto params_only;
 		
 		if (content_type)
 		    usage();
		 
 		p = strchr(type,'/');
		if (!p) {
		    
 		no_type:
 		    lib_error(CATGETS(elm_msg_cat, FastmailSet, 
 				      FastmailBadTypeont,
 				      "Bad type on on -t: %s"),
 			      optarg);
 		
 		    free(X);
 		    goto FAILURE;
 		}
 
 		subtype = p+1;
 
 		while (p > type && whitespace(*(p-1)))
 		    p--;
 		*p = '\0';
 		
 		if (!*type)
 		    goto no_type;
 
 		while (*subtype && whitespace(*subtype))
 		    subtype++;
 
 		if (!*subtype)
 		    goto no_type;
 
 		p = subtype + strlen(subtype);
 		while (p > subtype && whitespace(*(p-1)))
 		    p--;
 		*p = '\0';
 
 		content_type = give_media_type(type,subtype,1);
 		
 		if (!content_type)
 		    goto no_type;
 
 	    params_only:
 		while(NULL != (A =  mime_parse_content_opts(NULL,&walk))) {
 		    int len = strlen(A);
 
 		    if (0 == strncmp(A,"charset=",8) && len > 8) {
 			char * X = dequote_opt(A+8,len-8);
 
 			ctype_charset = MIME_name_to_charset(X,1);

 			free(X);
 
 		    } else {
 			ctype_params = safe_realloc(ctype_params,
 						    ( ctype_param_count+1) *
 						    sizeof( ctype_params[0]));
 			
 			ctype_params[ctype_param_count++] = safe_strdup(A);
 		    }
 		}
 	    }
 
 	    free(X);
 	}
	    break;

	case '?' : 
	    usage();
	}
    }	
    

    if (from_string) {
	if (! mailing_headers.from.addrs ||
	    0 == addr_list_item_count(mailing_headers.from.addrs)) {
	    
	    struct string * comment = new_string(display_charset);
	    struct string * s1 = NULL;
	    char *s = safe_strdup(username); /* Use username as surface addr */
	    char *x;
	    
	    char * Q = from_addr_literal(mailer_info);
	    
	    int pos = add_expanded_addr_(&mailing_headers.from,
					 Q,from_string,comment);
	    free(Q); Q = NULL;
	    
	    /* No fancy quotation stuff on textual form for editing */
	    for (x = s; *x; x++) {
		switch(*x) {
		case '"': case '\\': case '(': case ')': case ':':
		case '<': case '>':
		    *x = '_';
		    break;
		}
	    }
	    
	    s1 = new_string2(display_charset,s2us(s));
	    
	    add_textual_addr_(&mailing_headers.from,s1,pos,1);
	    free(s); s = NULL;
	    free_string(&s1);
	    
	    free_string(&comment);		    
	}
    }
    
    elm_sfprintf(version_buff, sizeof version_buff,
		 FRM("%s PL%s"), VERSION, PATCHLEVEL);
    
#ifdef DEBUG
    { 
	int d = panic_dprint("\n\
======================================================\n\
Debug output of the FASTMAIL program (version %s).\n",
			     version_buff);

	if (d >= 50) {
#if 0	
	    panic_dprint("WARNING: Edit manually out sensitive information from that file!\n");
    
	    lower_prompt("WARNING: Debug file may include passwords -- edit it!");
	    sleep(5+sleepmsg);	    
#endif
	}
    }
#endif

    if (optind >= argc) {
	usage();
    }
    
    filename = argv[optind++];

    if (optind >= argc) {
	usage();
    }

    if (0 == strcmp(filename, "-")) {
	char * tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir");
	if (!tmp)
	    tmp = "/tmp/";

	filename = tempname = elm_message(FRM("%selmfstmail%d"),tmp,getpid());

	F = safeopen_rdwr(filename);

	if (F) {
	    int ch;

	    if (-1 == unlink(filename)) {
		int err = errno;

		lib_error(CATGETS(elm_msg_cat, FastmailSet, 
				  FastmailUnlinkTmpFileError,
				  "Error to unlink temporary file %s: %s\n"),
			  filename,strerror(err));
	    }

	    while (EOF != (ch = getchar())) {
		
		if (EOF == putc(ch,F)) {
		    int err = errno;

		    lib_error(CATGETS(elm_msg_cat, FastmailSet, 
				      FastmailTmpFileError,
				      "Error to write temporary file %s: %s\n"),
			      filename,strerror(err));
		    
		    exitcode = 1;
		    goto FAILURE;
		}
	    }
	    
	    if (ferror(stdin)) {
		int err = errno;

		lib_error(CATGETS(elm_msg_cat, FastmailSet, 
				  FastmailStdInputError,
				  "Error when reading from standard input: %s\n"),
			  filename,strerror(err));
		
		exitcode = 1;
		goto FAILURE;

	    }

	    if (EOF == fflush(F)) {
		int err = errno;
		
		lib_error(CATGETS(elm_msg_cat, FastmailSet, 
				  FastmailTmpFileError,
				  "Error to write temporary file %s: %s\n"),
			  filename,strerror(err));
		
		exitcode = 1;
		goto FAILURE;
	    }			

	    rewind(F);
	}
	

    } else {
	F = fopen(filename,"r");
	ext = strrchr (filename, '.');
	if (ext)
	    ext++;

    }

    if (!F) {
	int err = errno;

	lib_error(CATGETS(elm_msg_cat, FastmailSet, 
			  FastmailCantFind,
			  "Error: can't find file %s: %s\n"), 
		  filename,strerror(err));

	exitcode = 1;
	goto FAILURE;
    }

    if (user_mimetypes_map)
	scanlist = get_scan_list(user_mimetypes_map,ext);

    if (system_mimetypes_map) {
	struct scan_list * X = get_scan_list(system_mimetypes_map,ext);
	
	if (X) {
	    if (scanlist) {
		append_scanlist(scanlist,X);
		free_scan_list(&X);
	    } else
		scanlist = X;
	}
    }
    
    
    { struct scan_list * Y = get_scan_list(builtin_mimetypes_map,ext);
    
	if (Y) {
	    if (scanlist) {
		append_scanlist(scanlist,Y);
		free_scan_list(&Y);
	    } else
		scanlist = Y;
	}
    }

    test_utf8 = is_utf8_charset(ctype_charset);

    need_enc = needs_encoding(F,scanlist,test_utf8);

    if (0 != (need_enc & HAVE_BINARY)) {

	if (query_mailer_info(mailer_info,MI_HAVE_BINARYMIME)) {
	    
	    DPRINT(Debug,4,(&Debug,"HAVE_BINARY: Mailer supports BINARYMIME\n"));
	    message_encoding = ENCODING_BINARY;
	} else if (an >= 2) {
	    DPRINT(Debug,4,(&Debug,"HAVE_BINARY: Just send binary anyway\n"));
	    message_encoding = ENCODING_BINARY;
	} else {
	    DPRINT(Debug,4,(&Debug,"HAVE_BINARY: Use Base64\n"));
	    message_encoding = ENCODING_BASE64;
	}

    } else if (0 != (need_enc & HAVE_8BIT)) {

	if (query_mailer_info(mailer_info,MI_HAVE_8BITMIME)) {
	    DPRINT(Debug,4,(&Debug,"HAVE_8BIT: Mailer supports 8BITMIME\n"));
	    message_encoding = ENCODING_8BIT;
	} else if (an >= 1) {
	    DPRINT(Debug,4,(&Debug,"HAVE_8BIT: Just send 8bit anyway\n"));
	    message_encoding = ENCODING_8BIT;
	} else {
	    DPRINT(Debug,4,(&Debug,"HAVE_8BIT: Use quoted-printable\n"));
	    message_encoding = ENCODING_QUOTED;
	}
	    
    }


    type_item = loc_mime_type_from_scan(scanlist);

    if (type_item) {
	const char * params = mime_type_to_params(type_item,NULL);

	content_type = mime_type_to_media_type(type_item);

	DPRINT(Debug,4,(&Debug,"Found type %s/%s\n",
			get_major_type_name(content_type),
			get_subtype_name(content_type)));

	if (params) {
	    char * X = safe_strdup(params);
	    char * walk = NULL;
	    char *A;
	    
	    for (A = mime_parse_content_opts(X,&walk); A;
		 A = mime_parse_content_opts(NULL,&walk)) {

		if (! check_8bit_str(A)) {
		    ctype_params = safe_realloc(ctype_params,
					       ( ctype_param_count+1) *
					       sizeof( ctype_params[0]));

		    ctype_params[ctype_param_count++] = safe_strdup(A);
		} else {
		    DPRINT(Debug,4,(&Debug,"Ignoring parameter %s\n",A));
		}
	    }
		     
	    free(X);
	}
    }

    if (! content_type &&
	0 != (need_enc & HAVE_BINARY) &&
	0 != (need_enc & HAVE_CTRL)) {

	/* If HAVE_BINARY is set, but HAVE_CTRL is not set, file 
	   have just long lines. So it still can be text.
	*/

	content_type = give_media_type2(MIME_TYPE_APPLICATION,
					"octet-stream", 1);

	if (content_type) {
	    DPRINT(Debug,4,(&Debug,"Binary data - setting type %s/%s\n",
			    get_major_type_name(content_type),
			    get_subtype_name(content_type)));
	}

    }

    /* Add charset paramater ? */
    if ((need_enc || ! charset_ok_p(ctype_charset)) 
	&&
	( ! content_type  ||
	  get_major_type_code(content_type) == MIME_TYPE_TEXT)) {

	if (! content_type) {
	    content_type = give_media_type2(MIME_TYPE_TEXT,"plain",0);

	    if (content_type) {
		DPRINT(Debug,4,(&Debug,"Setting type %s/%s\n",
				get_major_type_name(content_type),
				get_subtype_name(content_type)));
	    }
	}

	{
	    const char *ctype_MIME_name = get_charset_MIME_name(ctype_charset);

	    if (ctype_MIME_name) {
		if (istrcmp(ctype_MIME_name,"US-ASCII") == 0 &&
		    (need_enc & HAVE_8BIT)) {
		    
		    lib_error(CATGETS(elm_msg_cat, FastmailSet, FastmailUsingUNKNOWN8BIT,
				      "Text has 8BIT data and charset=US-ASCII, using charset=UNKNOWN-8BIT instead."));
		    ctype_params = safe_realloc(ctype_params,
						( ctype_param_count+1) *
						sizeof( ctype_params[0]));
		    
		    ctype_params[ctype_param_count++] = safe_strdup("charset=UNKNOWN-8BIT");
		    
		} else if (test_utf8 && (need_enc & HAVE_8BIT) &&
			   ! (need_enc & HAVE_UTF8)) {
		    
		    lib_error(CATGETS(elm_msg_cat, FastmailSet, FastmailNotUTF8,
				      "Text is not UTF-8 text, using charset=UNKNOWN-8BIT instead."));
		    ctype_params = safe_realloc(ctype_params,
						( ctype_param_count+1) *
						sizeof( ctype_params[0]));
		    
		    ctype_params[ctype_param_count++] = safe_strdup("charset=UNKNOWN-8BIT");
		    
		    
		} else {
		    ctype_params = safe_realloc(ctype_params,
						( ctype_param_count+1) *
						sizeof( ctype_params[0]));
		    
		    ctype_params[ctype_param_count++] = 
			elm_message(FRM("charset=%Q"),
				    ctype_MIME_name);
		    
		    DPRINT(Debug,4,(&Debug,"Setting charset %s\n",
				    ctype_MIME_name));
		
		}	
	    }
	}
    }


    if (content_type) {
	is_text = give_text_type_code(content_type);

	/* Content-Transfer-Encoding 7bit or 8bit implies text type */

	if (0 == is_text && message_encoding < ENCODING_BINARY) {

	    /* DO not use base65 for structured data (ie is_text < 0) */

	    if (query_mailer_info(mailer_info,MI_HAVE_BINARYMIME)) {
		
		DPRINT(Debug,4,(&Debug,
				"Non-text content type: Mailer supports BINARYMIME\n"));
		message_encoding = ENCODING_BINARY;
	    } else if (an >= 2) {
		DPRINT(Debug,4,(&Debug,"Non-text content type: Just send binary anyway\n"));
		message_encoding = ENCODING_BINARY;
	    } else if (need_enc) {
		DPRINT(Debug,4,(&Debug,"Non-text content type: Use Base64\n"));
		message_encoding = ENCODING_BASE64;
	    } else {
		DPRINT(Debug,4,(&Debug,"Non-text content type: Use quoted-printable\n"));
		message_encoding = ENCODING_QUOTED;
	    }
	}
    }
    
    while (optind < argc) {
	simple_add_addr(&mailing_headers.to,mailer_info,
			argv[optind++],system_charset,
			!allow_no_hdrencoding); 
    }


    addrs3 = argv_from_headers(&mailing_headers);

    M = mailer_init(addrs3,0,0,mailer_info,
		    mailing_headers.env_from );

    free_argv(&addrs3);

    if (!M) {

	exitcode = 1;
	goto FAILURE;
    }
 
    mailer = get_mail_outfd(M);

    if (ENCODING_BINARY == message_encoding)
	set_out_state_EOLN(mailer,1);

    /** Subject must appear even if "null" and must be first
	at top of headers for mail because the
	pure System V.3 mailer, in its infinite wisdom, now
	assumes that anything the user sends is part of the 
	message body unless either:
	1. the "-s" flag is used (although it doesn't seem
	   to be supported on all implementations?)
	2. the first line is "Subject:".  If so, then it'll
	   read until a blank line and assume all are meant
	   to be headers.
	So the gory solution here is to move the Subject: line
	up to the top.  I assume it won't break anyone elses program
	or anything anyway (besides, RFC-822 specifies that the *order*
	of headers is irrelevant).  Gahhhhh....

	If we have been configured for a smart mailer then we don't want
	to add a from line.  If the user has specified one then we have
	to honor their wishes.  If they've just given a 'from name' then
	we'll just put in the username and hope the mailer can add the
	correct domain in.
    **/

    if (mailing_headers.subject)
	write_string_header(mailer,"Subject",
			    mailing_headers.subject,
			    message_encoding,
			    !allow_no_hdrencoding,
			    display_charset            /* What is correct charset? */
			    );


    if (mailing_headers.from.addrs) 
	write_addr_header(mailer,"From",mailing_headers.from.addrs,
			  message_encoding,
			  !allow_no_hdrencoding,
			  display_charset      /* What is correct charset? */,
			  0
			  );

    write_text_header(mailer,"Date",get_arpa_date(),
		      message_encoding);
    
    if (mailing_headers.to.addrs) {
	int cnt;

	if (1 == (cnt = addr_list_item_count(mailing_headers.to.addrs))) {

	    int group;
	    const struct address * X = 
		addr_list_get_item(mailing_headers.to.addrs,0,&group);
	    const char * addr = address_get_ascii_addr(X);

	    elm_sfprintf(title, sizeof title,
			 CATGETS(elm_msg_cat, FastmailSet, FastmailMailTo,
				 "Mail to %.50s..."),
			 addr ? addr : "(someone)");
	} else
	    elm_sfprintf(title, sizeof title,
			 CATGETS(elm_msg_cat, FastmailSet, FastmailMailToXRecipients,
				 "Mail to %d recipients (To:)"),
			 cnt);

 	write_addr_header(mailer,"To",mailing_headers.to.addrs,
 			  message_encoding,
 			  !allow_no_hdrencoding,
 			  display_charset            /* What is correct charset? */,
			  1);

    } else
	elm_sfprintf(title, sizeof title,
		     CATGETS(elm_msg_cat, FastmailSet, FastmailMailToSomeone,
			     "Mail to someone..."));

    if (mailing_headers.reply_to.addrs)
	write_addr_header(mailer,"Reply-To",mailing_headers.reply_to.addrs,
			  message_encoding,
			  !allow_no_hdrencoding,
			  display_charset    /* What is correct charset? */,
			  1
			  );
	
    if (mailing_headers.cc.addrs)
	write_addr_header(mailer,"CC",mailing_headers.cc.addrs,
			  message_encoding,
			  !allow_no_hdrencoding,
			  display_charset    /* What is correct charset? */,
			  1
			  );

    
    if (mailing_headers.references)
	write_references_header(mailer,"References",
				mailing_headers.references,
				message_encoding,
				!allow_no_hdrencoding,
				display_charset            /* What is correct charset? */
				);

    if (mailing_headers.precedence)
	write_text_header(mailer,"Precedence",mailing_headers.precedence,
			  message_encoding);

    if (mailing_headers.user_header_count > 0) {
	int i;
	for (i = 0; i <  mailing_headers.user_header_count; i++) {
	    write_string_header(mailer,
				give_header_name(mailing_headers.user_header[i].
						 name),
				mailing_headers.user_header[i].value,
				message_encoding,
				!allow_no_hdrencoding,
				display_charset            /* What is correct charset? */
				);
	}
    }

    if (mailing_headers.in_reply_to)
	write_references_header(mailer,"In-Reply-To",
				mailing_headers.in_reply_to,
				message_encoding,
				!allow_no_hdrencoding,
				display_charset            /* What is correct charset? */
				);
    
    if (comments)
	write_string_header(mailer,"Comments",
			    comments,
			    message_encoding,
			    !allow_no_hdrencoding,
			    display_charset            /* What is correct charset? */
			    );
				
    write_text_header(mailer,"Mime-Version","1.0",message_encoding);
 
    if (message_encoding <  sizeof mime_encode_names / sizeof (mime_encode_names[0])
 	&&
	message_encoding > ENCODING_NONE) {
 	state_printf(mailer,FRM("Content-Transfer-Encoding: %s"),
		     mime_encode_names[message_encoding]);
	print_EOLN(mailer,message_encoding);
    }
 
    if (content_type) {    
 	state_printf(mailer,FRM("Content-Type: %s/%s"),
 		     get_major_type_name(content_type),
 		     get_subtype_name(content_type));
  	
 	if (ctype_params) {
 	    int i;
	    
 	    state_putc(';',mailer);
	    
 	    if (strlen(get_major_type_name(content_type)) +
 		strlen(get_subtype_name(content_type)) > 70)
 		print_EOLN(mailer,message_encoding);
	     	    
 	    for (i = 0; i < ctype_param_count; i++) {
 		
 		state_putc(' ',mailer);
 		state_puts(ctype_params[i],mailer);
 		
 		if (i < ctype_param_count-1) {
 		    state_putc(';',mailer);
		    print_EOLN(mailer,message_encoding);
 		}
 	    }	    
 	}
 	
 	print_EOLN(mailer,message_encoding);
    }
 	
#ifndef NO_XHEADER
    state_printf(mailer, FRM("X-Mailer: fastmail [version %s PL%s]\n"),
		 VERSION, PATCHLEVEL);
#endif /* !NO_XHEADER */

    print_EOLN(mailer,message_encoding);
    

    {
 	int r;
	
	
 	if (message_encoding <= ENCODING_BINARY &&
 	    message_encoding >= ENCODING_NONE) {
  
	    if (is_text > 0 || message_encoding < ENCODING_BINARY) {
		
		char buffer[VERY_LONG_STRING];
		int len;
		
		while ((len = mail_gets (buffer, sizeof buffer, F)) > 0) {
		    
		    state_convert_EOLN(buffer,&len,sizeof buffer,mailer);
		    
		    state_put(buffer, len, mailer);
		}
	    } else {
		int ch;
		
		if (is_text < 0) {
		    
		    lib_error(CATGETS(elm_msg_cat, FastmailSet, 
				      FastmailBinaryEncodingNotSupported,
				      "%s: Binary encoding is not supported for structured types."),
			      filename);
		    
		}
		
		while (EOF != (ch = getc(F))) {
		    state_putc(ch,mailer);
		    
		}
	    }
	    
	} else if (ENCODING_QUOTED == message_encoding) {
 	    quoted_printable_encode(F,mailer,is_text, message_encoding,0);
 	} else if (ENCODING_BASE64 == message_encoding) {
 	    base64_encode(F,mailer,is_text, message_encoding,0);
 	} else {
 	    DPRINT(Debug,4,(&Debug,"bad encoding %d\n",message_encoding));
 	    goto FAILURE;
 	}
	
 	if(ferror(F)) {
 	    int err = errno;
	    
 	    lib_error(CATGETS(elm_msg_cat, FastmailSet, 
 			      FastmailFileReadError,
 			      "Error when reading from file file %s: %s\n"), 
 		      filename,strerror(err));
	    
 	    exitcode = 1;
 	    goto FAILURE;
 	}
	    	

	r = mail_backend2(&M,mail_sent,
			  0,title,mailing_message);
  
 	DPRINT(Debug,4,(&Debug,"mail_backend2 returned %d\n",r));

 	/* 0   may be:    start failed
 	 *                mailing contined on backend
 	 *
  	 * 1              command completed, perhaps with failure 
 	 */
	
 	exitcode = 0;
    } 

 FAILURE:
    if (mailer_info)
	free_mailer_info(&mailer_info);

    free_mailing_headers(&mailing_headers);

    if (F)
 	fclose(F);
    
    if (tempname)
 	free(tempname);
    
    if (scanlist)
 	free_scan_list(&scanlist);
    
    if (ctype_params) {
 	int i;
	
 	for (i = 0; i < ctype_param_count; i++) 
 	    free(ctype_params[i]);
 	free(ctype_params);
    }
    
    DPRINT(Debug,4,(&Debug,"exitcode %d\n",exitcode));
    
    exit(exitcode);
    /*NOTREACHED*/
}

static void usage()
{

	lib_error(CATGETS(elm_msg_cat, FastmailSet, FastmailUsage,
"Usage: fastmail {args} [ filename | - ] address(es)\n\
   where {args} can be;\n\
\t-b bcc-list\n\t-c cc-list\n\t-d ...\n\
\t-C comments\n\t-f from-name\n\t-F from-addr\n\
\t-i msg-id\n\t-r reply-to\n\t-R references\n\
\t-P precedence\n\t-s subject\n"));

	lib_error(CATGETS(elm_msg_cat, FastmailSet, FastmailUsage2,
"\t-t type/subtype\n\t-t ;charset=character-set\n\
\t-t content-type\n\n"));

	exit(1);
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
